import tensorflow as tf

# ----------------------------------------------------------------------------------- #
#                                   KL-divergences                                    #
# ----------------------------------------------------------------------------------- #


def _mc_integral_kl(st_a, st_b, n_samp):
    samples = st_b.sample(sample_shape=n_samp)
    log_density_ratio = st_a.log_prob(samples) - st_b.log_prob(samples)
    density_a = st_a.prob(samples)
    return tf.reduce_mean(density_a * log_density_ratio, axis=0)


# TODO only register these if they aren't already so this module can be reloaded
# this is from here: https://mast.queensu.ca/~communications/Papers/GiAlLi13.pdf
@tf.distributions.RegisterKL(tf.distributions.Laplace, tf.distributions.Laplace)
def _kl_laplace(lap_i, lap_j, name=None):
    return tf.log(lap_j.scale / lap_i.scale) + tf.abs(lap_i.loc - lap_j.loc) / lap_j.scale + (lap_i.scale / lap_j.scale) * tf.exp(-tf.abs(lap_i.loc - lap_j.loc) / lap_i.scale) - 1


# monte carlo integral, could use a closed-form formula instead
@tf.distributions.RegisterKL(tf.distributions.StudentT, tf.distributions.StudentT)
def _kl_student(st_a, st_b, name=None):
    return _mc_integral_kl(st_a, st_b, 1)


# TODO there should be a closed-form solution for this
@tf.distributions.RegisterKL(tf.distributions.Normal, tf.distributions.Laplace)
def _kl_gauss_laplace(st_a, st_b, name=None):
    return _mc_integral_kl(st_a, st_b, 1)